# Copyright (c) 2017 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os

def expand_paths(paths, subpaths):
    results = []
    for path in paths:
        if path:
            for subpath in subpaths:
                full_path = os.path.join(path, subpath)
                if os.path.exists(full_path):
                    results.append(full_path)
    return results

def find_library(paths, name):
    paths = expand_paths(paths, ["", "lib", "lib64"])
    for path in paths:
        full_path = os.path.join(path, name)
        if os.path.exists(full_path):
            return full_path
    return name



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import pathlib
import shutil
import sys
import re

from rocm_docs import ROCmDocs

# We need to add the location of the rocrand Python module to the PATH
# in order to build the documentation of that module
docs_dir_path = pathlib.Path(__file__).parent
python_dir_path = docs_dir_path.parent / 'python' / 'rocrand'
sys.path.append(str(python_dir_path))

with open('../CMakeLists.txt', encoding='utf-8') as f:
    match = re.search(r'rocm_setup_version\( VERSION\s+\"?([0-9.]+)[^0-9.]+', f.read())
    if not match:
        raise ValueError("VERSION not found!")
    version_number = match[1]
left_nav_title = f"rocRAND {version_number} Documentation"
shutil.copy2('../library/src/fortran/README.md', './fortran-api-reference.md')

# for PDF output on Read the Docs
project = "rocRAND Documentation"
author = "Advanced Micro Devices, Inc."
copyright = "Copyright (c) 2025 Advanced Micro Devices, Inc. All rights reserved."
version = version_number
release = version_number

external_toc_path = "./sphinx/_toc.yml"
external_projects_current_project = "rocrand"

docs_core = ROCmDocs(left_nav_title)
docs_core.run_doxygen(doxygen_root="doxygen", doxygen_path="doxygen/xml")
docs_core.enable_api_reference()
docs_core.setup()

external_projects_current_project = "rocrand"

for sphinx_var in ROCmDocs.SPHINX_VARS:
    globals()[sphinx_var] = getattr(docs_core, sphinx_var)

# Suppresses "WARNING: toctree directive not expected with external-toc"
# Ideally suppression wouldn't be needed; see sphinx-external-toc#36
suppress_warnings = ["etoc.toctree"]

cpp_id_attributes = ["__forceinline__", "__device__", "__host__", "ROCRANDAPI"]



# Copyright (c) 2017-2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Minimal HIP wrapper"""

import os
import ctypes
import ctypes.util
from ctypes import *

import numbers
import numpy as np

from .utils import find_library

# Finalize is only supported by python >= 3.4
try:
  from weakref import finalize
except ImportError:
  from backports.weakref import finalize


class HipError(Exception):
    """Run-time HIP error."""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)

hipSuccess = 0
hipMemcpyDeviceToHost = 2

def check_hip(status):
    if status != hipSuccess:
        raise HipError(status)

hip = None

HIP_PATHS = [
    os.getenv("ROCM_PATH"),
    os.getenv("HIP_PATH"),
    "/opt/rocm",
    "/opt/rocm/hip"]

CUDA_PATHS = [
    os.getenv("CUDA_PATH"),
    "/opt/cuda"]

def load_hip():
    global hip

    loading_errors = []
    if hip is None:
        try:
            hip = CDLL(find_library(HIP_PATHS, "libamdhip64.so"), mode=RTLD_GLOBAL)
        except OSError as e:
            loading_errors.append(str(e))

    if hip is None:
        try:
            cuda = CDLL(find_library(CUDA_PATHS, "libcudart.so"), mode=RTLD_GLOBAL)
        except OSError as e:
            loading_errors.append(str(e))
        else:
            hip = cuda
            # Aliases
            hip.hipMalloc = cuda.cudaMalloc
            hip.hipFree = cuda.cudaFree
            hip.hipMemcpy = cuda.cudaMemcpy
            hip.hipStreamSynchronize = cuda.cudaStreamSynchronize

    if hip is None:
        raise ImportError("both libcudart.so and libamdhip64.so cannot be loaded: " +
                ", ".join(loading_errors))

class MemoryPointer(object):
    def __init__(self, nbytes):
        self.ptr = c_void_p()
        check_hip(hip.hipMalloc(byref(self.ptr), c_size_t(nbytes)))
        finalize(self, MemoryPointer._finalize, self.ptr)

    @classmethod
    def _finalize(cls, ptr):
        check_hip(hip.hipFree(ptr))

def device_pointer(dary):
    return dary.data.ptr

class DeviceNDArray(object):
    def __init__(self, shape, dtype, data=None):
        """Create an empty device-side array.

        This class is a limited version of :class:`numpy.ndarray` for device-side
        arrays.

        :param shape: Shape of the array (see :attr:`numpy.ndarray.shape`)
        :param dtype: Type of the array (see :attr:`numpy.ndarray.dtype`)
        :param data:  existing HIP device-side memory pointer
        """

        dtype = np.dtype(dtype)

        if isinstance(shape, numbers.Integral):
            shape = (shape,)
        shape = tuple(shape)

        size = np.prod(shape)

        self.shape = shape
        self.dtype = dtype
        self.size = size
        self.nbytes = self.dtype.itemsize * self.size

        if data is None:
            self.data = MemoryPointer(self.nbytes)
        else:
            self.data = data

    def copy_to_host(self, ary=None):
        """Copy from data device memory to host memory.

        If **ary** is passed then **ary** must have the same **dtype**
        and greater or equal **size** as **self** has.

        If **ary** is not passed then a new :class:`numpy.ndarray` will be
        created.

        :param ary: NumPy array (:class:`numpy.ndarray`)

        :returns: a new array or **ary**
        """
        if ary is None:
            ary = np.empty(self.shape, self.dtype)
        else:
            if self.dtype != ary.dtype:
                raise TypeError("self and ary must have the same dtype")
            if self.size > ary.size:
                raise ValueError("size of self must be less than size of ary")

        dst = ary.ctypes.data_as(c_void_p)
        src = device_pointer(self)
        check_hip(hip.hipMemcpy(dst, src, c_size_t(self.nbytes), hipMemcpyDeviceToHost))

        return ary

def empty(shape, dtype):
    """Create a new device-side array of given shape and type, without initializing entries.

    This function is a limited version of :func:`numpy.empty` for device-side
    arrays.

    Example::

        import rocrand
        import numpy as np

        gen = rocrand.QRNG(ndim=5)
        d_a = rocrand.empty((5, 10000), dtype=np.float32)
        gen.uniform(d_a)
        a = d_a.copy_to_host()
        print(a)

    See :class:`DeviceNDArray`

    :param shape: Shape of the array (see :attr:`numpy.ndarray.shape`)
    :type shape: int or tuple of int
    :param dtype: Type of the array (see :attr:`numpy.ndarray.dtype`)
    """
    return DeviceNDArray(shape, dtype)

def stream_synchronize(stream):
    """Blocks until all previously queued operations on the provided stream complete.

    :param stream: The HIP stream to synchronize.
    """
    check_hip(hip.hipStreamSynchronize(stream))



# Copyright (c) 2017 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import rocrand
import numpy as np

interactive = True
if not interactive:
    import matplotlib
    matplotlib.use("Agg")
import matplotlib.pyplot as plt


figsize = None
if not interactive:
    figsize = (30, 20)

plt.figure(1, figsize=figsize)
plt.suptitle("Estimating Pi using Monte Carlo method")

n = 10000

cols = 3
for col in range(cols):
    if col == 0:
        title = "numpy.random"
        np.random.seed(0)
        xy = np.random.uniform(size=(2, n))
    elif col == 1:
        title = "rocrand.PRNG"
        gen = rocrand.PRNG()
        xy = np.empty(shape=(2, n))
        gen.uniform(xy)
    elif col == 2:
        title = "rocrand.QRNG"
        gen = rocrand.QRNG(ndim=2)
        xy = np.empty(shape=(2, n))
        gen.uniform(xy)

    inside = xy[0]**2 + xy[1]**2 <= 1.0

    in_xy  = xy[:,  inside]
    out_xy = xy[:, ~inside]

    trials = np.arange(1, n + 1)
    hits = np.cumsum(inside)
    pis = 4.0 * (hits.astype(np.float) / trials)

    plt.subplot(2, 3, col + 1, aspect="equal")
    plt.title(title)
    plt.scatter(*in_xy,  c="g", marker=".", s=0.5)
    plt.scatter(*out_xy, c="r", marker=".", s=0.5)

    plt.subplot(2, 3, cols + col + 1)
    plt.axhline(y=np.pi, color="b", linestyle="-")
    plt.plot(trials, pis, color="c", linewidth=0.5)
    plt.ylim(2.8, 3.4)

plt.show()

if not interactive:
    plt.savefig("pi.png")



from setuptools import setup, find_packages
import sys


with open("README.md") as f:
    readme = f.read()

REQUIRED_PACKAGES = ["numpy"]

# Use a backport of weakref.finalizers if not supported by the standard library
if sys.version_info < (3, 4):
    REQUIRED_PACKAGES.append("backports.weakref")

version = "1.0"
release = "1.0.0"
setup(
    name="rocrand",
    version=release,
    description="rocRAND Python Wrapper",
    long_description=readme,
    author="Advanced Micro Devices, Inc.",
    url="https://github.com/ROCmSoftwarePlatform/rocRAND",
    license="MIT",
    packages=["rocrand"],
    install_requires=REQUIRED_PACKAGES,
    command_options={
        "build_sphinx": {
            "version": ("setup.py", version),
            "release": ("setup.py", release)}},
)



#!/usr/bin/env -S python3

from datetime import datetime
import json
import json5
import os
import re
import pandas
from pandas import DataFrame
import seaborn as sns
from matplotlib import pyplot as plt
import argparse
from jinja2 import Environment, PackageLoader

env = Environment(
    loader=PackageLoader("select_best_config"),
    lstrip_blocks=True,
    trim_blocks=True
)


def load_default_configs_json(in_dir: str) -> dict:
    """
    Reads the config_defaults.json object for rocRAND's generators and returns them as dict.
    """

    with open(os.path.join(in_dir, "config_defaults.json")) as f:
        json_data = json5.load(f)

    return json_data

def load_benchmark_result_json(path: str) -> DataFrame:
    """
    Reads and processes a JSON file produced by rocRAND's tuning benchmark.
    """
    def guess_gcn_architecture(path: str, raw_json_data: dict):
        arch_regex = re.compile(r'gfx[0-9a-f]+')
        arch_name = raw_json_data['context']['hdp_gcn_arch_name']
        _, filename = os.path.split(path)
        for name in (arch_name, filename):
            gfx_match = re.search(arch_regex, name)
            if gfx_match:
                return gfx_match.group(0)
        return 'unknown'

    with open(path) as f:
        raw_json_data = json.load(f)
    benchmark_data = DataFrame(raw_json_data['benchmarks'])

    gb_per_s_series = benchmark_data['bytes_per_second'] / 1000 / 1000 / 1000
    gb_per_s_series.name = 'gb_per_s'

    # Extract the groups of the regex match to a DataFrame
    
    name_regex = r'^(?P<generator>\S+?)_(?P<distribution>uniform|normal|log_normal|poisson)_(?P<value_type>(unsigned_)?(int|short|char|long_long|float|half|double))_t(?P<block_size>\d+)_b(?P<grid_size>\d+)'

    extracted_data = benchmark_data['name'].str.extract(name_regex)
    extracted_data['block_size'] = extracted_data['block_size'].astype(int)
    extracted_data['grid_size']  = extracted_data['grid_size'].astype(int)

    # Merge the regex matches and the gb_per_s series
    benchmark_data = pandas.concat([extracted_data, gb_per_s_series], axis=1)

    # Figure out architecture
    benchmark_data['arch'] = guess_gcn_architecture(path, raw_json_data)
    return benchmark_data


def get_best_config_for_arch(benchmark_data: DataFrame, default_configs: dict):
    """
    Calculates the best config for each architecture, that on average provides the highest
    performance across generated types and distributions.
    """
    def config_to_goodness(subf: DataFrame) -> pandas.Series:
        relative_perfs = subf['normalized_perf'].reset_index(
            drop=True)
        # Averages the performance over the different distribution/value_types
        # to get a single average performance per configuration of the generator
        return relative_perfs.sum() / relative_perfs.count()

    best_configs = {}
    for gen, generator_df in benchmark_data.groupby('generator'):
        best_config_for_generator = {}
        default_block_size = default_configs[gen]['block_size']
        default_grid_size  = default_configs[gen]['grid_size']

        for arch, arch_df in generator_df.groupby('arch'):

            temp_df = DataFrame()
            # Normalize other configurations wrt default config per distribution/value_type
            for _, perf_df in arch_df.groupby(['distribution', 'value_type']):
                default_config_perf = perf_df.loc[
                    (perf_df['block_size'] == default_block_size) &
                    (perf_df['grid_size']  == default_grid_size)
                ].gb_per_s.iloc[0]
                perf_df['normalized_perf'] = perf_df.gb_per_s / default_config_perf
                temp_df = pandas.concat([temp_df, perf_df])

            arch_df = temp_df
            config_goodness = arch_df.groupby(
                ['block_size', 'grid_size']).apply(config_to_goodness)
            config_goodness = config_goodness.sort_values(ascending=False)
            config_goodness = config_goodness.to_frame(name='normalized_perf').reset_index()

            # config_goodness is sorted by normalized_perf, so just take the first element
            best_block_size = config_goodness.iloc[0]['block_size'].astype(int)
            best_grid_size  = config_goodness.iloc[0]['grid_size'].astype(int)
            best_perf       = config_goodness.iloc[0]['normalized_perf']

            # check for performance regressions and possible alternative configurations
            detailed_performance = arch_df[(arch_df['block_size'] == best_block_size) &
                                           (arch_df['grid_size']  == best_grid_size)]
            if (detailed_performance['normalized_perf'] < 0.98).any():
                print("WARNING: configuration with best average performance has performance regressions for some distributions:")
                print("Average speedup: ", best_perf)
                print(detailed_performance.to_string())
                # set index to config for easier looping
                config_goodness = config_goodness.set_index(['block_size', 'grid_size'])
                other_configs = config_goodness[(config_goodness['normalized_perf'] > 1.0) &
                                                (config_goodness['normalized_perf'] != best_perf)]
                if len(other_configs.index) == 0:
                    print("\nNo other configs available that provide an average speedup over the default config!\n")
                # check other configs that have an average speedup for regressions
                found_config_without_regressions = False
                for config, row in other_configs.iterrows():
                    block_size, grid_size = config
                    cur_detailed_perf = arch_df[(arch_df['block_size'] == block_size) &
                                                (arch_df['grid_size']  == grid_size)]
                    if (cur_detailed_perf['normalized_perf'] > 1.0).all():
                        print("Next best config that provides an average speedup over the default config without regressions:")
                        print("Average speedup: ", row['normalized_perf'])
                        print(cur_detailed_perf.to_string())
                        print("\n")
                        found_config_without_regressions = True
                        break
                if not found_config_without_regressions:
                    print("No other config provides a speedup over the default config without any regressions for certain distributions!\n")

            best_config_for_generator[arch] = {
                'block_size': best_block_size, 'grid_size': best_grid_size}
            generator_df.loc[(generator_df['generator'] == gen) & (generator_df['arch'] == arch)] = arch_df
        best_configs[gen] = best_config_for_generator
    return best_configs


def plot_benchmark_data_for_all_arches(benchmark_data: DataFrame, out_path: str):
    """
    Plots comparative figures of the benchmark results for each architecture.
    """
    def plot_benchmark_results(benchmark_data: DataFrame, arch: str, gen: str, out_path: str):
        def draw_heatmap(*args, **kwargs):
            data: DataFrame = kwargs.pop('data')
            d = data.pivot_table(
                index=args[1], columns=args[0], values=args[2])
            sns.heatmap(d, **kwargs)

        grid = sns.FacetGrid(benchmark_data, col='benchmark', col_wrap=3)
        grid.map_dataframe(draw_heatmap, 'grid_size_', 'block_size_', 'gb_per_s',
                           cbar=False, square=True, annot=True, fmt='.2f', cmap='coolwarm', linewidth=1)
        fig = plt.gcf()
        fig.suptitle(f'rocRAND benchmarks ({gen} on {arch}, GB/s)')
        num_grid_sizes = benchmark_data['grid_size'].unique().astype(
            bool).sum()
        fig.set_size_inches(16/6*num_grid_sizes, 18)
        fig.savefig(out_path)
        plt.close()

    benchmark_data['block_size_'] = benchmark_data['block_size'].astype(str).str.pad(
        4, 'left', ' ')
    benchmark_data['grid_size_'] = benchmark_data['grid_size'].astype(str).str.pad(
        4, 'left', ' ')
    benchmark_data['benchmark'] = benchmark_data['value_type'].astype(
        str) + ',' + benchmark_data['distribution'].astype(str)
    base_path, full_filename = os.path.split(out_path)
    filename, ext = os.path.splitext(full_filename)
    for arch in benchmark_data['arch'].unique():
        arch_benchmark_data = benchmark_data[benchmark_data['arch'] == arch]
        for gen in arch_benchmark_data['generator'].unique():
            gen_benchmark_data = arch_benchmark_data[arch_benchmark_data['generator'] == gen]
            plot_benchmark_results(gen_benchmark_data, arch, gen, os.path.join(base_path, f'{filename}_{arch}_{gen}{ext}'))


def generate_config_files(out_dir: str, best_data: dict):
    def generator_type(generator_name: str) -> str:
        if 'sobol' in generator_name.lower():
            return 'quasi'
        return 'pseudo'

    year = datetime.now().year
    algorithm_template = env.get_template("config_template")
    for generator, best_configs in best_data.items():
        with open(os.path.join(out_dir,  f"{generator}_config.hpp"), "w") as text_file:
            text_file.write(algorithm_template.render(
                generator=generator, generator_type=generator_type(generator), configs=best_configs, year=year))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('json', nargs='+')
    parser.add_argument('--plot-out', help="Output directory for plots. Plots will only be generated, when this argument is provided. Generating these can take quite some while.")
    parser.add_argument('--out-dir')
    parser.add_argument('--default-config-dir',
                        default=os.path.dirname(os.path.realpath(__file__)),
                        help="Path to directory of 'config_defaults.json'")
    args = parser.parse_args()

    default_configs = load_default_configs_json(args.default_config_dir)
    benchmark = pandas.concat(
        [load_benchmark_result_json(j) for j in args.json], ignore_index=True)
    best_configs = get_best_config_for_arch(benchmark, default_configs)
    if args.plot_out:
        plot_benchmark_data_for_all_arches(benchmark, args.plot_out)
    if args.out_dir:
        generate_config_files(args.out_dir, best_configs)



# Copyright (c) 2017-2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""rocRAND Python Wrapper"""

import os
import ctypes
import ctypes.util
from ctypes import *

import numbers
import numpy as np

from .hip import load_hip, HIP_PATHS, stream_synchronize
from .hip import empty, DeviceNDArray, device_pointer

from .utils import find_library, expand_paths

# Finalize is only supported by python >= 3.4
try:
  from weakref import finalize
except ImportError:
  from backports.weakref import finalize


ROCRAND_PATHS = [
        os.getenv("ROCRAND_PATH")
    ] + expand_paths(HIP_PATHS, ["", "rocrand"])

def load_rocrand():
    global rocrand

    try:
        rocrand = CDLL(find_library(ROCRAND_PATHS, "librocrand.so"))
    except OSError as e:
        raise ImportError("librocrand.so cannot be loaded: " + str(e))

    load_hip()

# Delay the loading of rocrand to the first use
# so no code is executed when loading this module
class _load_rocrand_on_access(object):
    def __getattribute__(self, name):
        global rocrand
        load_rocrand()
        return getattr(rocrand, name)

rocrand = _load_rocrand_on_access()

ROCRAND_RNG_PSEUDO_DEFAULT = 400
ROCRAND_RNG_PSEUDO_XORWOW = 401
ROCRAND_RNG_PSEUDO_MRG32K3A = 402
ROCRAND_RNG_PSEUDO_MTGP32 = 403
ROCRAND_RNG_PSEUDO_PHILOX4_32_10 = 404
ROCRAND_RNG_PSEUDO_MRG31K3P = 405
ROCRAND_RNG_PSEUDO_LFSR113 = 406
ROCRAND_RNG_PSEUDO_MT19937 = 407
ROCRAND_RNG_PSEUDO_THREEFRY2_32_20 = 408
ROCRAND_RNG_PSEUDO_THREEFRY2_64_20 = 409
ROCRAND_RNG_PSEUDO_THREEFRY4_32_20 = 410
ROCRAND_RNG_PSEUDO_THREEFRY4_64_20 = 411
ROCRAND_RNG_QUASI_DEFAULT = 500
ROCRAND_RNG_QUASI_SOBOL32 = 501
ROCRAND_RNG_QUASI_SCRAMBLED_SOBOL32 = 502
ROCRAND_RNG_QUASI_SOBOL64 = 504
ROCRAND_RNG_QUASI_SCRAMBLED_SOBOL64 = 505

ROCRAND_STATUS_SUCCESS = 0
ROCRAND_STATUS_VERSION_MISMATCH = 100
ROCRAND_STATUS_NOT_CREATED = 101
ROCRAND_STATUS_ALLOCATION_FAILED = 102
ROCRAND_STATUS_TYPE_ERROR = 103
ROCRAND_STATUS_OUT_OF_RANGE = 104
ROCRAND_STATUS_LENGTH_NOT_MULTIPLE = 105
ROCRAND_STATUS_DOUBLE_PRECISION_REQUIRED = 106
ROCRAND_STATUS_LAUNCH_FAILURE = 107
ROCRAND_STATUS_INTERNAL_ERROR = 108

ROCRAND_STATUS = {
    ROCRAND_STATUS_SUCCESS: (
        "ROCRAND_STATUS_SUCCESS",
        "Success"),
    ROCRAND_STATUS_VERSION_MISMATCH: (
        "ROCRAND_STATUS_VERSION_MISMATCH",
        "Header file and linked library version do not match"),
    ROCRAND_STATUS_NOT_CREATED: (
        "ROCRAND_STATUS_NOT_CREATED",
        "Generator was not created using rocrand_create_generator"),
    ROCRAND_STATUS_ALLOCATION_FAILED: (
        "ROCRAND_STATUS_ALLOCATION_FAILED",
        "Memory allocation failed during execution"),
    ROCRAND_STATUS_TYPE_ERROR: (
        "ROCRAND_STATUS_TYPE_ERROR",
        "Generator type is wrong"),
    ROCRAND_STATUS_OUT_OF_RANGE: (
        "ROCRAND_STATUS_OUT_OF_RANGE",
        "Argument out of range"),
    ROCRAND_STATUS_LENGTH_NOT_MULTIPLE: (
        "ROCRAND_STATUS_LENGTH_NOT_MULTIPLE",
        "Length requested is not a multiple of dimension"),
    ROCRAND_STATUS_DOUBLE_PRECISION_REQUIRED: (
        "ROCRAND_STATUS_DOUBLE_PRECISION_REQUIRED",
        "GPU does not have double precision"),
    ROCRAND_STATUS_LAUNCH_FAILURE: (
        "ROCRAND_STATUS_LAUNCH_FAILURE",
        "Kernel launch failure"),
    ROCRAND_STATUS_INTERNAL_ERROR: (
        "ROCRAND_STATUS_INTERNAL_ERROR",
        "Internal library error")
}

def check_rocrand(status):
    if status != ROCRAND_STATUS_SUCCESS:
        raise RocRandError(status)


class RocRandError(Exception):
    """Run-time rocRAND error."""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        if self.value in ROCRAND_STATUS:
            v, s = ROCRAND_STATUS[self.value]
        else:
            v, s = str(self.value), "Unknown error"
        return "{} ({})".format(s, v)


class RNG(object):
    """Random number generator base class."""

    def __init__(self, rngtype, offset=None, stream=None, is_host=False):
        self._gen = c_void_p()
        create_fun = rocrand.rocrand_create_generator_host if is_host else rocrand.rocrand_create_generator
        check_rocrand(create_fun(byref(self._gen), rngtype))
        finalize(self, RNG._finalize, self._gen)

        self._offset = 0
        if offset is not None:
            self.offset = offset

        self._stream = None
        if stream is not None:
            self.stream = stream
        self.is_host = is_host

    @classmethod
    def _finalize(cls, gen):
        check_rocrand(rocrand.rocrand_destroy_generator(gen))

    @property
    def offset(self):
        """Mutable attribute of the offset of random numbers sequence.

        Setting this attribute resets the sequence.
        """
        return self._offset

    @offset.setter
    def offset(self, offset):
        """Mutable attribute of HIP stream for all kernel launches of the generator.

        All functions will use this stream.
        *None* means default stream.
        """
        check_rocrand(rocrand.rocrand_set_offset(self._gen, c_ulonglong(offset)))
        self._offset = offset

    @property
    def stream(self):
        return self._stream

    @stream.setter
    def stream(self, stream):
        check_rocrand(rocrand.rocrand_set_stream(self._gen, stream))
        self._stream = stream

    def _generate_device(self, gen_func, ary, size, *args):
        if size is None:
            size = ary.size
        elif size > ary.size:
            raise ValueError("requested size is greater than ary")

        if isinstance(ary, np.ndarray):
            dary, needs_conversion = empty(size, ary.dtype), True
        elif isinstance(ary, DeviceNDArray):
            dary, needs_conversion = ary, False
        else:
            raise TypeError("unsupported type {}".format(type(ary)))

        check_rocrand(gen_func(self._gen, device_pointer(dary), c_size_t(size), *args))

        if needs_conversion:
            dary.copy_to_host(ary)

    def _generate_host(self, gen_func, ary, size, *args):
        if size is None:
            size = ary.size
        elif size > ary.size:
            raise ValueError("requested size is greater than ary")

        if isinstance(ary, DeviceNDArray):
            raise TypeError("Generate called with a device-side array on a host-side generator. "
                            "For device arrays, instantiate a device-side generator")
        elif not isinstance(ary, np.ndarray):
            raise TypeError("unsupported type {}".format(type(ary)))

        check_rocrand(gen_func(self._gen, ctypes.c_void_p(ary.ctypes.data), c_size_t(size), *args))
        # If we're generating on the host, synchronize the kernel to match the
        # behavior of a device-side generator being called with a host array (_generate_device).
        stream_synchronize(self.stream)

    def _generate(self, gen_func, ary, size, *args):
        if self.is_host:
            self._generate_host(gen_func, ary, size, *args)
        else:
            self._generate_device(gen_func, ary, size, *args)

    def generate(self, ary, size=None):
        """Generates uniformly distributed integers.

        Generates **size** (if present) or **ary.size** uniformly distributed
        integers and saves them to **ary**.

        Supported **dtype** of **ary**: :class:`numpy.uint32`, :class:`numpy.int32`.

        :param ary:  NumPy array (:class:`numpy.ndarray`) or
                     HIP device-side array (:class:`DeviceNDArray`)
        :param size: Number of samples to generate, default to **ary.size**
        """
        if ary.dtype in (np.uint32, np.int32):
            self._generate(
                rocrand.rocrand_generate,
                ary, size)
        else:
            raise TypeError("unsupported type {}".format(ary.dtype))

    def uniform(self, ary, size=None):
        """Generates uniformly distributed floats.

        Generates **size** (if present) or **ary.size** uniformly distributed
        floats and saves them to **ary**.

        Supported **dtype** of **ary**: :class:`numpy.float32`, :class:`numpy.float64`.

        Generated numbers are between 0.0 and 1.0, excluding 0.0 and
        including 1.0.

        :param ary:  NumPy array (:class:`numpy.ndarray`) or
                     HIP device-side array (:class:`DeviceNDArray`)
        :param size: Number of samples to generate, default to **ary.size**
        """
        if ary.dtype == np.float32:
            self._generate(
                rocrand.rocrand_generate_uniform,
                ary, size)
        elif ary.dtype == np.float64:
            self._generate(
                rocrand.rocrand_generate_uniform_double,
                ary, size)
        else:
            raise TypeError("unsupported type {}".format(ary.dtype))

    def normal(self, ary, mean, stddev, size=None):
        """Generates normally distributed floats.

        Generates **size** (if present) or **ary.size** normally distributed
        floats and saves them to **ary**.

        Supported **dtype** of **ary**: :class:`numpy.float32`, :class:`numpy.float64`.

        :param ary:    NumPy array (:class:`numpy.ndarray`) or
                       HIP device-side array (:class:`DeviceNDArray`)
        :param mean:   Mean value of normal distribution
        :param stddev: Standard deviation value of normal distribution
        :param size:   Number of samples to generate, default to **ary.size**
        """
        if ary.dtype == np.float32:
            self._generate(
                rocrand.rocrand_generate_normal,
                ary, size,
                c_float(mean), c_float(stddev))
        elif ary.dtype == np.float64:
            self._generate(
                rocrand.rocrand_generate_normal_double,
                ary, size,
                c_double(mean), c_double(stddev))
        else:
            raise TypeError("unsupported type {}".format(ary.dtype))

    def lognormal(self, ary, mean, stddev, size=None):
        """Generates log-normally distributed floats.

        Generates **size** (if present) or **ary.size** log-normally distributed
        floats and saves them to **ary**.

        Supported **dtype** of **ary**: :class:`numpy.float32`, :class:`numpy.float64`.

        :param ary:    NumPy array (:class:`numpy.ndarray`) or
                       HIP device-side array (:class:`DeviceNDArray`)
        :param mean:   Mean value of log normal distribution
        :param stddev: Standard deviation value of log normal distribution
        :param size:   Number of samples to generate, default to **ary.size**
        """
        if ary.dtype == np.float32:
            self._generate(
                rocrand.rocrand_generate_log_normal,
                ary, size,
                c_float(mean), c_float(stddev))
        elif ary.dtype == np.float64:
            self._generate(
                rocrand.rocrand_generate_log_normal_double,
                ary, size,
                c_double(mean), c_double(stddev))
        else:
            raise TypeError("unsupported type {}".format(ary.dtype))

    def poisson(self, ary, lmbd, size=None):
        """Generates Poisson-distributed integers.

        Generates **size** (if present) or **ary.size** Poisson-distributed
        integers and saves them to **ary**.

        Supported **dtype** of **ary**: :class:`numpy.uint32`, :class:`numpy.int32`.

        :param ary:   NumPy array (:class:`numpy.ndarray`) or
                      HIP device-side array (:class:`DeviceNDArray`)
        :param lmbd:  lambda for the Poisson distribution
        :param size:  Number of samples to generate, default to **ary.size**
        """
        if ary.dtype in (np.uint32, np.int32):
            self._generate(
                rocrand.rocrand_generate_poisson,
                ary, size,
                c_double(lmbd))
        else:
            raise TypeError("unsupported type {}".format(ary.dtype))


class PRNG(RNG):
    DEFAULT         = ROCRAND_RNG_PSEUDO_DEFAULT
    """Default pseudo-random generator type, :const:`XORWOW`"""
    XORWOW          = ROCRAND_RNG_PSEUDO_XORWOW
    """XORWOW pseudo-random generator type"""
    MRG31K3P        = ROCRAND_RNG_PSEUDO_MRG31K3P
    """MRG31k3p pseudo-random generator type"""
    MRG32K3A        = ROCRAND_RNG_PSEUDO_MRG32K3A
    """MRG32k3a pseudo-random generator type"""
    MTGP32          = ROCRAND_RNG_PSEUDO_MTGP32
    """Mersenne Twister MTGP32 pseudo-random generator type"""
    MT19937         = ROCRAND_RNG_PSEUDO_MT19937
    """Mersenne Twister pseudo-random generator type"""
    PHILOX4_32_10   = ROCRAND_RNG_PSEUDO_PHILOX4_32_10
    """PHILOX_4x32 (10 rounds) pseudo-random generator type"""
    LFSR113         = ROCRAND_RNG_PSEUDO_LFSR113
    """LFSR113 pseudo-random generator type"""
    THREEFRY2_32_20 = ROCRAND_RNG_PSEUDO_THREEFRY2_32_20
    """THREEFRY2_32_20 pseudo-random generator type"""
    THREEFRY2_64_20 = ROCRAND_RNG_PSEUDO_THREEFRY2_64_20
    """THREEFRY2_64_20 pseudo-random generator type"""
    THREEFRY4_32_20 = ROCRAND_RNG_PSEUDO_THREEFRY4_32_20
    """THREEFRY4_32_20 pseudo-random generator type"""
    THREEFRY4_64_20 = ROCRAND_RNG_PSEUDO_THREEFRY4_64_20
    """THREEFRY4_64_20 pseudo-random generator type"""

    def __init__(self, rngtype=DEFAULT, seed=None, offset=None, stream=None, is_host=False):
        """Creates a new pseudo-random number generator.

        A new pseudo-random number generator of type **rngtype** is initialized
        with given **seed**, **offset** and **stream**. If **is_host** is set to true, the generator
        operates on the host CPU and memory instead on the device. Note, that not all generators
        may support the host option.

        Values of **rngtype**:

        * :const:`DEFAULT`
        * :const:`XORWOW`
        * :const:`MRG31K3P`
        * :const:`MRG32K3A`
        * :const:`MTGP32`
        * :const:`MT19937`
        * :const:`PHILOX4_32_10`
        * :const:`LFSR113`
        * :const:`THREEFRY2_32_20`
        * :const:`THREEFRY2_64_20`
        * :const:`THREEFRY4_32_20`
        * :const:`THREEFRY4_64_20`

        :param rngtype: Type of pseudo-random number generator to create
        :param seed:    Initial seed value
        :param offset:  Initial offset of random numbers sequence
        :param stream:  HIP stream for all kernel launches of the generator
        :param is_host:    If True, the generator operates on the host CPU, not on the device

        Example::

            import rocrand
            import numpy as np

            gen = rocrand.PRNG(rocrand.PRNG.PHILOX4_32_10, seed=123456)
            a = np.empty(1000, np.int32)
            gen.poisson(a, 10.0)
            print(a)
        """
        super(PRNG, self).__init__(rngtype, offset=offset, stream=stream, is_host=is_host)

        self._seed = None
        if seed is not None:
            self.seed = seed

    @property
    def seed(self):
        """Mutable attribute of the seed of random numbers sequence.

        Setting this attribute resets the sequence.
        """
        return self._seed

    @seed.setter
    def seed(self, seed):
        check_rocrand(rocrand.rocrand_set_seed(self._gen, c_ulonglong(seed)))
        self._seed = seed


class QRNG(RNG):
    DEFAULT           = ROCRAND_RNG_QUASI_DEFAULT
    """Default quasi-random generator type, :const:`SOBOL32`"""
    SOBOL32           = ROCRAND_RNG_QUASI_SOBOL32
    """Sobol32 quasi-random generator type"""
    SCRAMBLED_SOBOL32           = ROCRAND_RNG_QUASI_SCRAMBLED_SOBOL32
    """Scrambled Sobol32 quasi-random generator type"""
    SOBOL64           = ROCRAND_RNG_QUASI_SOBOL64
    """Sobol64 quasi-random generator type"""
    SCRAMBLED_SOBOL64           = ROCRAND_RNG_QUASI_SCRAMBLED_SOBOL64
    """Scrambled Sobol64 quasi-random generator type"""

    def __init__(self, rngtype=DEFAULT, ndim=None, offset=None, stream=None, is_host=False):
        """Creates a new quasi-random number generator.

        A new quasi-random number generator of type **rngtype** is initialized
        with given **ndim**, **offset** and **stream**. If **is_host** is set to true, the generator
        operates on the host CPU and memory instead on the device. Note, that not all generators
        may support the host option.

        Values of **rngtype**:

        * :const:`DEFAULT`
        * :const:`SOBOL32`
        * :const:`SCRAMBLED_SOBOL32`
        * :const:`SOBOL64`
        * :const:`SCRAMBLED_SOBOL64`

        Values of **ndim** are 1 to 20000.

        :param rngtype: Type of quasi-random number generator to create
        :param ndim:    Number of dimensions
        :param offset:  Initial offset of random numbers sequence
        :param stream:  HIP stream for all kernel launches of the generator
        :param is_host:    If True, the generator operates on the host CPU, not on the device

        Example::

            import rocrand
            import numpy as np

            gen = rocrand.QRNG(rocrand.QRNG.SOBOL32, ndim=4)
            a = np.empty(1000, np.float32)
            gen.normal(a, 0.0, 1.0)
            print(a)
        """

        super(QRNG, self).__init__(rngtype, offset=offset, stream=stream, is_host=is_host)

        self._ndim = 1
        if ndim is not None:
            self.ndim = ndim

    @property
    def ndim(self):
        """Mutable attribute of the number of dimensions of random numbers sequence.

        Supported values are 1 to 20000.
        Setting this attribute resets the sequence.
        """
        return self._ndim

    @ndim.setter
    def ndim(self, ndim):
        check_rocrand(rocrand.rocrand_set_quasi_random_generator_dimensions(self._gen, c_uint(ndim)))
        self._ndim = ndim


def get_version():
    """Returns the version number of the rocRAND library."""
    version = c_int(0)
    check_rocrand(rocrand.rocrand_get_version(byref(version)))
    return version.value



#! /usr/bin/env python3

import re
import subprocess


def get_grid_sizes(rocminfo_output: str) -> str:
    match = re.search(r'^\s*Name:\s*gfx\d+.*?^\s*Compute Unit:\s*(\d+)',
                      rocminfo_output, flags=re.MULTILINE | re.DOTALL)
    if not match:
        raise Exception('Could not find Compute Unit info in rocminfo output')
    num_compute_units = int(match.group(1))
    compute_unit_multipliers = [4, 5, 8, 10, 16, 32]
    min_grid_size = 128
    max_grid_size = 4096

    grid_sizes = [128, 256, 512, 1024, 2048]
    for cu_mul in compute_unit_multipliers:
        new_grid_size = cu_mul * num_compute_units
        if new_grid_size >= min_grid_size and new_grid_size <= max_grid_size:
            grid_sizes.append(new_grid_size)
    grid_sizes = list(set(grid_sizes))  # Unique
    grid_sizes.sort()
    return ', '.join(str(i) for i in grid_sizes)


if __name__ == '__main__':
    rocminfo_out = subprocess.check_output('rocminfo', encoding='utf-8')
    print(get_grid_sizes(rocminfo_out))



# Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import unittest
import math
import numpy as np

from rocrand import *


def make_test(BaseClass, subname, **kwargs):
    class Class(BaseClass):
        def setUp(self):
            super(Class, self).setUp()

    name = BaseClass.__name__ + subname
    globals()[name] = type(name, (Class,), kwargs)

class TestVersion(unittest.TestCase):
    def test_version(self):
        self.assertGreater(get_version(), 0)

class TestCtor(unittest.TestCase):
    def test_ctor(self):
        with self.assertRaises(RocRandError):
            PRNG(1234)
        with self.assertRaises(RocRandError):
            QRNG(1234)

class TestRNGBase(unittest.TestCase):
    rngtype = None

    def setUp(self):
        if not self.rngtype:
            self.skipTest("rngtype is not set")

class TestCtorPRNG(TestRNGBase):
    def test_ctor(self):
        PRNG(self.rngtype)
        PRNG(self.rngtype, seed=123456)
        PRNG(self.rngtype, offset=987654)
        PRNG(self.rngtype, seed=2345678, offset=7654)

make_test(TestCtorPRNG, "DEFAULT",         rngtype=PRNG.DEFAULT)
make_test(TestCtorPRNG, "LFSR113",         rngtype=PRNG.LFSR113)
make_test(TestCtorPRNG, "XORWOW",          rngtype=PRNG.XORWOW)
make_test(TestCtorPRNG, "MRG31K3P",        rngtype=PRNG.MRG31K3P)
make_test(TestCtorPRNG, "MRG32K3A",        rngtype=PRNG.MRG32K3A)
make_test(TestCtorPRNG, "PHILOX4_32_10",   rngtype=PRNG.PHILOX4_32_10)
make_test(TestCtorPRNG, "THREEFRY2_32_20", rngtype=PRNG.THREEFRY2_32_20)
make_test(TestCtorPRNG, "THREEFRY2_64_20", rngtype=PRNG.THREEFRY2_64_20)
make_test(TestCtorPRNG, "THREEFRY4_32_20", rngtype=PRNG.THREEFRY4_32_20)
make_test(TestCtorPRNG, "THREEFRY4_64_20", rngtype=PRNG.THREEFRY4_64_20)

class TestCtorPRNGMT(TestRNGBase):
    def test_ctor(self):
        PRNG(self.rngtype)
        PRNG(self.rngtype, seed=123456)
        with self.assertRaises(RocRandError):
            PRNG(self.rngtype, offset=987654)
        with self.assertRaises(RocRandError):
            PRNG(self.rngtype, seed=2345678, offset=7654)

make_test(TestCtorPRNGMT, "MTGP32",  rngtype=PRNG.MTGP32)
make_test(TestCtorPRNGMT, "MT19937", rngtype=PRNG.MT19937)

class TestCtorQRNG(TestRNGBase):
    def test_ctor(self):
        QRNG(self.rngtype)
        QRNG(self.rngtype, ndim=10)
        QRNG(self.rngtype, offset=987654)
        QRNG(self.rngtype, ndim=123, offset=7654)

make_test(TestCtorQRNG, "DEFAULT", rngtype=QRNG.DEFAULT)
make_test(TestCtorQRNG, "SOBOL32", rngtype=QRNG.SOBOL32)
make_test(TestCtorQRNG, "SCRAMBLED_SOBOL32", rngtype=QRNG.SCRAMBLED_SOBOL32)
make_test(TestCtorQRNG, "SOBOL64", rngtype=QRNG.SOBOL64)
make_test(TestCtorQRNG, "SCRAMBLED_SOBOL64", rngtype=QRNG.SCRAMBLED_SOBOL64)

class TestParamsPRNG(TestRNGBase):
    def setUp(self):
        super(TestParamsPRNG, self).setUp()
        self.rng = PRNG(self.rngtype)

    def tearDown(self):
        del self.rng

    def test_seed(self):
        self.assertIsNone(self.rng.seed)
        self.rng.seed = 0
        self.assertEqual(self.rng.seed, 0)
        self.rng.seed = 54654634456365
        self.assertEqual(self.rng.seed, 54654634456365)

    def test_offset(self):
        self.assertEqual(self.rng.offset, 0)
        self.rng.offset = 0
        self.assertEqual(self.rng.offset, 0)
        self.rng.offset = 2323423
        self.assertEqual(self.rng.offset, 2323423)

make_test(TestParamsPRNG, "DEFAULT",         rngtype=PRNG.DEFAULT)
make_test(TestParamsPRNG, "LFSR113",         rngtype=PRNG.LFSR113)
make_test(TestParamsPRNG, "XORWOW",          rngtype=PRNG.XORWOW)
make_test(TestParamsPRNG, "MRG31K3P",        rngtype=PRNG.MRG31K3P)
make_test(TestParamsPRNG, "MRG32K3A",        rngtype=PRNG.MRG32K3A)
make_test(TestParamsPRNG, "PHILOX4_32_10",   rngtype=PRNG.PHILOX4_32_10)
make_test(TestParamsPRNG, "THREEFRY2_32_20", rngtype=PRNG.THREEFRY2_32_20)
make_test(TestParamsPRNG, "THREEFRY2_64_20", rngtype=PRNG.THREEFRY2_64_20)
make_test(TestParamsPRNG, "THREEFRY4_32_20", rngtype=PRNG.THREEFRY4_32_20)
make_test(TestParamsPRNG, "THREEFRY4_64_20", rngtype=PRNG.THREEFRY4_64_20)

class TestParamsPRNGMT(TestRNGBase):
    def setUp(self):
        super(TestParamsPRNGMT, self).setUp()
        self.rng = PRNG(self.rngtype)

    def tearDown(self):
        del self.rng

    def test_seed(self):
        self.assertIsNone(self.rng.seed)
        self.rng.seed = 0
        self.assertEqual(self.rng.seed, 0)
        self.rng.seed = 54654634456365
        self.assertEqual(self.rng.seed, 54654634456365)

    def test_offset(self):
        self.assertEqual(self.rng.offset, 0)
        with self.assertRaises(RocRandError):
            self.rng.offset = 2323423

make_test(TestParamsPRNGMT, "MTGP32",  rngtype=PRNG.MTGP32)
make_test(TestParamsPRNGMT, "MT19937", rngtype=PRNG.MT19937)

class TestParamsQRNG(TestRNGBase):
    def setUp(self):
        super(TestParamsQRNG, self).setUp()
        self.rng = QRNG(self.rngtype)

    def tearDown(self):
        del self.rng

    def test_ndim(self):
        self.assertEqual(self.rng.ndim, 1)
        self.rng.ndim = 10
        self.assertEqual(self.rng.ndim, 10)
        self.rng.ndim = 123
        self.assertEqual(self.rng.ndim, 123)
        with self.assertRaises(RocRandError):
            self.rng.ndim = 0
        self.assertEqual(self.rng.ndim, 123)
        with self.assertRaises(RocRandError):
            self.rng.ndim = 30000
        self.assertEqual(self.rng.ndim, 123)

    def test_offset(self):
        self.assertEqual(self.rng.offset, 0)
        self.rng.offset = 0
        self.assertEqual(self.rng.offset, 0)
        self.rng.offset = 2323423
        self.assertEqual(self.rng.offset, 2323423)

make_test(TestParamsQRNG, "DEFAULT", rngtype=QRNG.DEFAULT)
make_test(TestParamsQRNG, "SOBOL32", rngtype=QRNG.SOBOL32)
make_test(TestParamsQRNG, "SCRAMBLED_SOBOL32", rngtype=QRNG.SCRAMBLED_SOBOL32)
make_test(TestParamsQRNG, "SOBOL64", rngtype=QRNG.SOBOL64)
make_test(TestParamsQRNG, "SCRAMBLED_SOBOL64", rngtype=QRNG.SCRAMBLED_SOBOL64)

OUTPUT_SIZE = 8192

class TestGenerate(TestRNGBase):
    def setUp(self):
        super(TestGenerate, self).setUp()
        self.rng = self.klass(self.rngtype, is_host=self.is_host)

    def tearDown(self):
        del self.rng

    def test_types(self):
        with self.assertRaises(TypeError):
            self.rng.uniform(np.empty(100, np.int8))
        with self.assertRaises(TypeError):
            self.rng.normal(np.empty(100, np.int8), 0.0, 1.0)
        with self.assertRaises(TypeError):
            self.rng.lognormal(np.empty(100, np.int8), 0.0, 1.0)
        with self.assertRaises(TypeError):
            self.rng.poisson(np.empty(100, np.float32), 100.0)

        self.rng.generate(np.empty(100, np.uint32))
        self.rng.generate(np.empty((10, 100), np.uint32))

        this_empty = np.empty if self.is_host else empty
        self.rng.uniform(this_empty(100, np.float32))
        self.rng.uniform(this_empty((10, 100), np.float32))

    def test_generate_uint32(self):
        output = np.empty(OUTPUT_SIZE, np.uint32)
        self.rng.generate(output)

        output = output.astype(np.float64)
        output /= 4294967295.0

        self.assertAlmostEqual(output.mean(), 0.5, delta=0.2)
        self.assertAlmostEqual(output.std(), pow(1 / 12.0, 0.5), delta=0.2 * pow(1 / 12.0, 0.5))

    def test_generate_int32(self):
        output = np.empty(OUTPUT_SIZE, np.int32)
        self.rng.generate(output)

        output = output.astype(np.float64)
        output /= 4294967295.0

        self.assertAlmostEqual(output.mean(), 0.0, delta=0.2)
        self.assertAlmostEqual(output.std(), pow(1 / 12.0, 0.5), delta=0.2 * pow(1 / 12.0, 0.5))

    def _test_uniform(self, dtype):
        output = np.empty(OUTPUT_SIZE, dtype)
        self.rng.uniform(output)

        self.assertAlmostEqual(output.mean(), 0.5, delta=0.2)
        self.assertAlmostEqual(output.std(), math.sqrt(1 / 12.0), delta=0.2 * math.sqrt(1 / 12.0))

    def test_uniform_float(self):
        self._test_uniform(np.float32)

    def test_uniform_double(self):
        self._test_uniform(np.float64)

    def _test_normal_real(self, dtype):
        output = np.empty(OUTPUT_SIZE, dtype)
        self.rng.normal(output, 0.0, 1.0)

        self.assertAlmostEqual(output.mean(), 0.0, delta=0.2)
        self.assertAlmostEqual(output.std(), 1.0, delta=0.2)

    def test_normal_float(self):
        self._test_normal_real(np.float32)

    def test_normal_double(self):
        self._test_normal_real(np.float64)

    def _test_lognormal_real(self, dtype):
        output = np.empty(OUTPUT_SIZE, dtype)
        self.rng.lognormal(output, 1.6, 0.25)

        mean = output.mean()
        stddev = output.std()
        logmean = math.log(mean * mean / math.sqrt(stddev + mean * mean))
        logstd = math.sqrt(math.log(1.0 + stddev / (mean * mean)))

        self.assertAlmostEqual(logmean, 1.6, delta=1.6 * 0.2)
        self.assertAlmostEqual(logstd, 0.25, delta=0.25 * 0.2)

    def test_lognormal_float(self):
        self._test_lognormal_real(np.float32)

    def test_lognormal_double(self):
        self._test_lognormal_real(np.float64)

    def test_poisson(self):
        if self.rngtype == PRNG.THREEFRY2_64_20 or self.rngtype == PRNG.THREEFRY4_64_20:
            self.skipTest("64-bits Threefry cannot generate 32-bits Poisson numbers")
        for lambda_value in [1.0, 5.5, 20.0, 100.0, 1234.5, 5000.0]:
            output = np.empty(OUTPUT_SIZE, np.uint32)
            self.rng.poisson(output, lambda_value)

            self.assertAlmostEqual(output.mean(), lambda_value, delta=max(1.0, lambda_value * 1e-1))
            self.assertAlmostEqual(output.var(), lambda_value, delta=max(1.0, lambda_value * 1e-1))

    def test_size(self):
        output = np.full(OUTPUT_SIZE * 2, 10.0, dtype=np.float64)
        self.rng.uniform(output, size=OUTPUT_SIZE)

        self.assertTrue((output[:OUTPUT_SIZE] <= 1.0).all())
        self.assertTrue((output[OUTPUT_SIZE:] == 10.0).all())

make_test(TestGenerate, "PRNG" + "DEFAULT",                klass=PRNG, rngtype=PRNG.DEFAULT,            is_host=False)
make_test(TestGenerate, "PRNG" + "XORWOW",                 klass=PRNG, rngtype=PRNG.XORWOW,             is_host=False)
make_test(TestGenerate, "PRNG" + "MRG31K3P",               klass=PRNG, rngtype=PRNG.MRG31K3P,           is_host=False)
make_test(TestGenerate, "PRNG" + "MRG31K3P_host",          klass=PRNG, rngtype=PRNG.MRG31K3P,           is_host=True)
make_test(TestGenerate, "PRNG" + "MRG32K3A",               klass=PRNG, rngtype=PRNG.MRG32K3A,           is_host=False)
make_test(TestGenerate, "PRNG" + "MTGP32",                 klass=PRNG, rngtype=PRNG.MTGP32,             is_host=False)
make_test(TestGenerate, "PRNG" + "MT19937",                klass=PRNG, rngtype=PRNG.MT19937,            is_host=False)
make_test(TestGenerate, "PRNG" + "PHILOX4_32_10",          klass=PRNG, rngtype=PRNG.PHILOX4_32_10,      is_host=False)
make_test(TestGenerate, "PRNG" + "PHILOX4_32_10_host",     klass=PRNG, rngtype=PRNG.PHILOX4_32_10,      is_host=True)
make_test(TestGenerate, "QRNG" + "DEFAULT",                klass=QRNG, rngtype=QRNG.DEFAULT,            is_host=False)
make_test(TestGenerate, "QRNG" + "SOBOL32",                klass=QRNG, rngtype=QRNG.SOBOL32,            is_host=False)
make_test(TestGenerate, "QRNG" + "SOBOL32_host",           klass=QRNG, rngtype=QRNG.SOBOL32,            is_host=True)
make_test(TestGenerate, "QRNG" + "SCRAMBLED_SOBOL32",      klass=QRNG, rngtype=QRNG.SCRAMBLED_SOBOL32,  is_host=False)
make_test(TestGenerate, "QRNG" + "SCRAMBLED_SOBOL32_host", klass=QRNG, rngtype=QRNG.SCRAMBLED_SOBOL32,  is_host=True)
make_test(TestGenerate, "QRNG" + "SOBOL64",                klass=QRNG, rngtype=QRNG.SOBOL64,            is_host=False)
make_test(TestGenerate, "QRNG" + "SOBOL64_host",           klass=QRNG, rngtype=QRNG.SOBOL64,            is_host=True)
make_test(TestGenerate, "QRNG" + "SCRAMBLED_SOBOL64",      klass=QRNG, rngtype=QRNG.SCRAMBLED_SOBOL64,  is_host=False)
make_test(TestGenerate, "QRNG" + "SCRAMBLED_SOBOL64_host", klass=QRNG, rngtype=QRNG.SCRAMBLED_SOBOL64,  is_host=True)
make_test(TestGenerate, "PRNG" + "LFSR113",                klass=PRNG, rngtype=PRNG.LFSR113,            is_host=False)
make_test(TestGenerate, "PRNG" + "THREEFRY2_32_20",        klass=PRNG, rngtype=PRNG.THREEFRY2_32_20,    is_host=False)
make_test(TestGenerate, "PRNG" + "THREEFRY2_64_20",        klass=PRNG, rngtype=PRNG.THREEFRY2_64_20,    is_host=False)
make_test(TestGenerate, "PRNG" + "THREEFRY4_32_20",        klass=PRNG, rngtype=PRNG.THREEFRY4_32_20,    is_host=False)
make_test(TestGenerate, "PRNG" + "THREEFRY4_64_20",        klass=PRNG, rngtype=PRNG.THREEFRY4_64_20,    is_host=False)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3
"""Copyright 2020-2023 Advanced Micro Devices, Inc.  All rights reserved.
Manage build and installation"""

import re
import sys
import os
import subprocess
import argparse
import ctypes
import pathlib
from fnmatch import fnmatchcase
import shutil

args = {}
param = {}
OS_info = {}

def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(description="""
    Checks build arguments
    """)

    default_gpus = 'gfx906:xnack-,gfx1030,gfx1100,gfx1101,gfx1102,gfx1151,gfx1200,gfx1201'

    parser.add_argument('-g', '--debug', required=False, default=False,  action='store_true',
                        help='Generate Debug build (default: False)')
    parser.add_argument(      '--build_dir', type=str, required=False, default="build",
                        help='Build directory path (default: build)')
    parser.add_argument(      '--deps_dir', type=str, required=False, default=None,
                        help='Dependencies directory path (default: build/deps)')
    parser.add_argument(      '--skip_ld_conf_entry', required=False, default=False)
    parser.add_argument(      '--static', required=False, default=False, dest='static_lib', action='store_true',
                        help='Generate static library build (default: False)')
    parser.add_argument('-c', '--clients', required=False, default=False, dest='build_clients', action='store_true',
                        help='Generate all client builds (default: False)')
    parser.add_argument('-p', '--package', required=False, default=False, dest='package', action='store_true',
                        help='Generate packages (default: False)')
    parser.add_argument('-i', '--install', required=False, default=False, dest='install', action='store_true',
                        help='Install after build (default: False)')
    parser.add_argument(      '--cmake-darg', required=False, dest='cmake_dargs', action='append', default=[],
                        help='List of additional cmake defines for builds (e.g. CMAKE_CXX_COMPILER_LAUNCHER=ccache)')
    parser.add_argument('-a', '--architecture', dest='gpu_architecture', required=False, default=default_gpus, #:sramecc+:xnack-" ) #gfx1030" ) #gfx906" ) # gfx1030" )
                        help='Set GPU architectures, e.g. all, gfx000, gfx803, gfx906:xnack-;gfx1030;gfx1100 (optional, default: all)')
    parser.add_argument('-v', '--verbose', required=False, default=False, action='store_true',
                        help='Verbose build (default: False)')
    return parser.parse_args()

def os_detect():
    inf_file = "/etc/os-release"
    if os.path.exists(inf_file):
        with open(inf_file) as f:
            for line in f:
                if "=" in line:
                    k,v = line.strip().split("=")
                    OS_info[k] = v.replace('"','')
    else:
        OS_info["ID"] = 'windows'
        OS_info["VERSION_ID"] = 10
    OS_info["NUM_PROC"] = os.cpu_count()
    print(OS_info)

def create_dir(dir_path):
    if os.path.isabs(dir_path):
        full_path = dir_path
    else:
        fullpath = os.path.join( os.getcwd(), dir_path )
    pathlib.Path(fullpath).mkdir(parents=True, exist_ok=True)
    return

def delete_dir(dir_path) :
    if (not os.path.exists(dir_path)):
        return
    if (OS_info["ID"] == 'windows'):
        run_cmd( "RMDIR" , f"/S /Q {dir_path}")
    else:
        linux_path = pathlib.Path(dir_path).absolute()
        #print( linux_path )
        run_cmd( "rm" , f"-rf {linux_path}")

def cmake_path(os_path):
    if OS_info["ID"] == "windows":
        return os_path.replace("\\", "/")
    else:
        return os.path.realpath(os_path)  
        
def config_cmd():
    global args
    global OS_info
    cwd_path = os.getcwd()
    src_path = cwd_path.replace("\\", "/")

    print( f"***************** {src_path}")
    cmake_executable = ""
    cmake_options = []
    cmake_platform_opts = []
    
    if (OS_info["ID"] == 'windows'):
        # we don't have ROCM on windows but have hip, ROCM can be downloaded if required
        raw_rocm_path = cmake_path(os.getenv('HIP_PATH', "C:/hip"))
        rocm_path = f'"{raw_rocm_path}"' # guard against spaces in path
        cmake_executable = "cmake.exe"
        toolchain = os.path.join( src_path, "toolchain-windows.cmake" )
        #set CPACK_PACKAGING_INSTALL_PREFIX= defined as blank as it is appended to end of path for archive creation
        cmake_platform_opts.append( f"-DCPACK_PACKAGING_INSTALL_PREFIX=" )
        cmake_platform_opts.append( f"-DCMAKE_INSTALL_PREFIX=\"C:/hipSDK\"" )
        generator = f"-G Ninja"
        # "-G \"Visual Studio 16 2019\" -A x64"  #  -G NMake ")  #
        cmake_options.append( generator )
    else:
        rocm_path = os.getenv( 'ROCM_PATH', "/opt/rocm")

        cmake_executable = 'cmake'
        if shutil.which('cmake3') is not None:
            cmake_executable = 'cmake3'

        toolchain = "toolchain-linux.cmake"
        cmake_platform_opts = [f"-DROCM_DIR:PATH={rocm_path}", f"-DCPACK_PACKAGING_INSTALL_PREFIX={rocm_path}"]

    tools = f"-DCMAKE_TOOLCHAIN_FILE={toolchain}"
    cmake_options.append( tools )

    cmake_options.extend( cmake_platform_opts)


  # build type
    cmake_config = ""
    build_dir = args.build_dir
    if not args.debug:
        build_path = os.path.join(build_dir, "release")
        cmake_config="Release"
    else:
        build_path = os.path.join(build_dir, "debug")
        cmake_config="Debug"

    cmake_options.append( f"-DCMAKE_BUILD_TYPE={cmake_config}" ) #--build {build_path}" )

    if args.deps_dir is None:
        deps_dir = os.path.abspath(os.path.join(build_dir, 'deps')).replace('\\','/')
    else:
        deps_dir = args.deps_dir
    cmake_base_options = f"-DROCM_PATH={rocm_path} -DCMAKE_PREFIX_PATH:PATH={rocm_path} -Drocrand_EXPORTS=1"
    cmake_options.append( cmake_base_options )

    print( cmake_options )

    # clean
    delete_dir( build_path )

    create_dir( os.path.join(build_path, "clients") )
    os.chdir( build_path )

    if args.static_lib:
        cmake_options.append( f"-DBUILD_SHARED_LIBS=OFF" )

    if args.skip_ld_conf_entry:
        cmake_options.append( f"-DROCM_DISABLE_LDCONFIG=ON" )

    if args.build_clients:
        cmake_options.append( f"-DBUILD_TEST=ON -DBUILD_BENCHMARK=ON -DBUILD_DIR={build_dir}" )

    cmake_options.append( f"-DAMDGPU_TARGETS={args.gpu_architecture}" )

    if args.cmake_dargs:
        for i in args.cmake_dargs:
          cmake_options.append( f"-D{i}" )

    cmake_options.append( f"{src_path}")

#   case "${ID}" in
#     centos|rhel)
#     cmake_options="${cmake_options} -DCMAKE_FIND_ROOT_PATH=/usr/lib64/llvm7.0/lib/cmake/"
#     ;;
#     windows)
#     cmake_options="${cmake_options} -DROCM_PATH=${rocm_path} -DROCM_DIR:PATH=${rocm_path} -DCMAKE_PREFIX_PATH:PATH=${rocm_path}"
#     cmake_options="${cmake_options} --debug-trycompile -DCMAKE_MAKE_PROGRAM=nmake.exe -DCMAKE_TOOLCHAIN_FILE=toolchain-windows.cmake"
#     # -G '"NMake Makefiles JOM"'"
#     ;;
#   esac
    cmd_opts = " ".join(cmake_options)

    return cmake_executable, cmd_opts


def make_cmd():
    global args
    global OS_info

    make_options = []

    if (OS_info["ID"] == 'windows'):
        make_executable = "cmake.exe --build ." # ninja"
        if args.verbose:
          make_options.append( "--verbose" )
        make_options.append( "--target all" )
        if args.install or args.package:
          make_options.append( "--target package" )
        if args.install:
            make_options.append( "--target install" )
    else:
        nproc = OS_info["NUM_PROC"]
        make_executable = f"make -j{nproc}"
        if args.verbose:
          make_options.append( "VERBOSE=1" )
        if args.install:
          make_options.append( "install" )
    cmd_opts = " ".join(make_options)

    return make_executable, cmd_opts

def run_cmd(exe, opts):
    program = f"{exe} {opts}"
    if sys.platform.startswith('win'):
        sh = True
    else:
        sh = True
    print(program)
    proc = subprocess.run(program, check=True, stderr=subprocess.STDOUT, shell=sh)
    #proc = subprocess.Popen(cmd, cwd=os.getcwd())
    #cwd=os.path.join(workingdir,"..",".."), stdout=fout, stderr=fout,
     #                       env=os.environ.copy())
    #proc.wait()
    return proc.returncode

def main():
    global args
    os_detect()
    args = parse_args()
    # configure
    exe, opts = config_cmd()
    run_cmd(exe, opts)

    # make/build/install
    exe, opts = make_cmd()
    run_cmd(exe, opts)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import json5 as json #json5 supports comments in json files
import os
import argparse
from jinja2 import Environment, PackageLoader

env = Environment(loader=PackageLoader("generate_config_defaults"))

def load_default_configs_json(in_dir: str) -> dict:
    """
    Reads the config_defaults.json object for rocRAND's generators,
    converts their name to the corresponding rocrand_rng_type and returns them as dict.
    """

    with open(os.path.join(in_dir, "config_defaults.json")) as f:
        json_data = json.load(f)

    def generator_type(generator_name: str) -> str:
        if 'SOBOL' in generator_name:
            return 'QUASI'
        return 'PSEUDO'
    
    def get_rocrand_rng_type(generator_name: str) -> str:
        return "ROCRAND_RNG_" + generator_type(generator_name) + "_" + generator_name

    default_configs = {}
    for generator_name, configs in json_data.items():
        generator_rocrand_enum = get_rocrand_rng_type(generator_name.upper())
        default_configs[generator_rocrand_enum] = configs

    return default_configs

def generate_config_file(out_dir: str, default_configs: dict) -> None:
    """
    Generates config_defaults.hpp from the config_defaults.json file.
    """

    algorithm_template = env.get_template("config_defaults_template")

    with open(os.path.join(out_dir,  f"config_defaults.hpp"), "w") as text_file:
        text_file.write(algorithm_template.render(configs=default_configs))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--in-dir',
        default=os.path.dirname(os.path.realpath(__file__)),
        help="Path to the directory that contains 'config_defaults.json',"
             "usually located in 'scripts/config-tuning/'")
    parser.add_argument('--out-dir', default=".")
    args = parser.parse_args()

    default_configs = load_default_configs_json(args.in_dir)
    generate_config_file(args.out_dir, default_configs)



# Copyright 2021 Advanced Micro Devices, Inc.
# This conanfile is used to install development requirements,
# e.g.
#   conan install -o clients=True -if build/deps .

from conans import ConanFile, CMake

class ConanPkgReqs(ConanFile):
    settings = "os", "compiler", "build_type", "arch"
    generators = "cmake_find_package"
    options = {
        "shared": [True, False],
        "clients": [True, False],
    }
    default_options = {
        "shared": True,
        "clients": False,
    }

    def requirements(self):
        if self.options.clients:
            self.requires("gtest/1.11.0")
